# -*- coding: utf-8 -*-

__uri__ = "https://pylick.readthedocs.io"
__author__ = "Nicola Borghi, Michele Moresco, Salvatore Quai"
__email__ = "nicola.borghi6@unibo.it,  michele.moresco@unibo.it"
__license__ = "GPLv3"
__description__ = "Python tool to measure spectral features in galaxy spectra"


from .__version__ import __version__  # isort:skip
from . import analysis
from . import indices
from . import io
from . import measurement
from . import plot

__all__ = [
    "analysis",
    "indices",
    "io",
    "measurement",
    "plot",
]



import os
import pylick

__ROOT__    = os.path.dirname(pylick.__file__)

__table__   = __ROOT__ + '/data/index_table.dat'
__style__   = __ROOT__ + '/data/plotbelli.style'

__dirRes__  = '../outputs/'
__dirPlot__ = '../outputs/plots/'



import matplotlib.pyplot as plt 
import matplotlib as mpl 
import numpy as np

from pylick.__config__ import __style__, __dirPlot__




def _default_settings(user_plParams):
    
    default_rcParams = {'lines.linewidth': 1.,
                        'axes.labelsize': 16.,
                        'xtick.top': True, 
                        'ytick.right': True,
                        'xtick.direction': 'in',
                        'ytick.direction': 'in',
                        'xtick.major.size': 5.,
                        'ytick.major.size': 5.,
                        'xtick.major.width': 1.,
                        'ytick.major.width': 1.,
                        'xtick.labelsize': 14.,
                        'ytick.labelsize': 14.,
                       }

    default_plParams = {'figsize': (8,4.8),
                        'xlab': r'Restframe wavelength [$\AA$]',
                        'ylab': r'Flux [unitless]',
                        # Colors of: flux, ferr, indices
                        'spec_colors': ['#000080', '#8E8EC2', (.4,.4,.4,1)],
                        # Style of the index patches (1: vband, 2: box)
                        'ind_style': 1,
                        # Fontsizes of: title, labels, indices 
                        'spec_fontsizes': [14, 14, 13],
                        'format': '.pdf',
                        # Debug
                        'inspect': False,
                        # Filename.ext of the plot to be saved
                        'filename': None,
                       }

    for k, v in default_rcParams.items():
        mpl.rcParams[k] = v

    for def_k, def_v in default_plParams.items():
        if not def_k in user_plParams:
            user_plParams[def_k] = def_v

    return user_plParams
        

def _rgba2rgb(rgba, background=(255,255,255)):  # Not used
    """Flattens transparent colors.

    Args:
        rgba (ndarray): 4D (RGBA) color array containing data with `float` type
        background (tuple, optional): White background color. Defaults to (255,255,255).

    Returns:
        ndarray: 3D (RGB) color array
    """
    row, col, ch    = rgba.shape
    rgb             = np.zeros((row, col, 3), dtype=float)
    r, g, b, a      = rgba[:,:,0], rgba[:,:,1], rgba[:,:,2], rgba[:,:,3]
    a               = np.asarray(a, dtype=float) / 255.0
    R, G, B         = background
    
    rgb[:,:,0], rgb[:,:,1], rgb[:,:,2] = r * a + (1.0 - a) * R, g * a + (1.0 - a) * G, b * a + (1.0 - a) * B

    return np.asarray(rgb, dtype='uint8')



def ax_addindex(ax, text, style, bbox_index, y_text, color, y_text_va="bottom", fontsize=13):
    """Plot and annotate an index region in data coordinates.

    Args:
        ax (axis): axis object
        text (str): name of the index to be displayed
        bbox_index (float, tuple): bounding box for the patch: (left, right, None, None) for vertical shadings, (left, right, down, up) for boxes
        y_text (float): y location of the text
        color (str, tuple): (r,g,b) color for the shadings. Must be in rgb format.
        fontsize (int): fontsize of the index to be displayed
    """ 

    if style==1:
        # Vertical shadings
        l, r = bbox_index[0], bbox_index[1]
        ax.axvspan(l, r, alpha=0.3, color=color, lw=1, zorder=4)
        text = ax.annotate(text, (0.5*(l+r), y_text), xycoords="data", 
                           rotation=90, ha='center',va=y_text_va, fontsize=fontsize, zorder=5)

    elif style==2:
        # Boxes
        l, r, d, u             = bbox_index
        alpha_face, alpha_edge = 0.2, 0.8
        col_face, col_edge     = [*color,alpha_face], [*color,alpha_edge] #darkdarkrblue
   
        rect = mpl.patches.Rectangle((l,d), width=(r-l), height=(u-d), 
                                     angle=0.0, facecolor=col_face, edgecolor=col_edge, lw=.7, ls='-')

        text = ax.annotate(text, (0.5*(l+r), y_text), xycoords="data", 
                           rotation=90, ha='center',va=y_text_va, fontsize=fontsize, zorder=5)
    else:            
        print("ERROR: Index style not suported: define bbox_index=(left, right, None, None) or (left, right, down, up)")

    return text


def spectrum_simple(wave, flux, ferr=None, mask=None, ax=None, settings={}):
    """
    Plots spetrum and associated errors (if given). A large error (9e9) is assigned to masked pixels, 
    making them appear as vertical lines. Additional settings may be passed as a dictionary using `settings`.

    Args:
        wave (np.ndarray): wavelength sampling of the spectrum
        flux (np.ndarray): spectral fluxes corresponding to `spec_wave`
        ferr (np.ndarray, optional): spectral fluxes' uncertainties.  Defaults to None.
        mask (np.ndarray, optional): boolean array flagging bad pixels to interpolate over. Defaults to None.
        ax (~matplotlib.axes.Axes, optional): overplot onto the provided ax object, otherwise create new figure. Defaults to None.
        settings (dict, optional): further instructions for the plot code. Defaults to {}.

    Returns:
        ~matplotlib.figure.Figure: Figure

    """

    # Load default configuration
    settings = _default_settings(settings)
    
    if ax is None:
        fig, ax = plt.subplots(figsize=settings["figsize"])

    if mask is not None:
        ferr[mask] = 9e9

    if ferr is not None:
        ax.fill_between(wave, flux-ferr, flux+ferr, step='mid', color=settings['spec_colors'][1], zorder=3)

    ax.step(wave, flux, where='mid', color=settings['spec_colors'][0], zorder=4)

    ax.set_xlabel(settings['xlab'], fontsize=settings['spec_fontsizes'][1])
    ax.set_ylabel(settings['ylab'], fontsize=settings['spec_fontsizes'][1])
    ax.set_ylim(np.nanmin(flux)*0.8, np.nanmax(flux)*1.05)

    return fig



def spectrum_with_indices(wave, flux, index_regions, index_names, index_units,
                          z=None, ferr=None, mask=None, ax=None, index_done=None, settings={}):

    """
    Plots :py:func:`spectrum_simple` with indices overlayed.

    Args:
        wave (np.ndarray): wavelength sampling of the spectrum
        flux (np.ndarray): spectral fluxes corresponding to `spec_wave`
        ferr (np.ndarray, optional): spectral fluxes' uncertainties.  Defaults to None.
        mask (np.ndarray, optional): boolean array flagging bad pixels to interpolate over. Defaults to None.
        ax (~matplotlib.axes.Axes, optional): overplot onto the provided ax object, otherwise create new figure. Defaults to None.
        settings (dict, optional): further instructions for the plot code. Defaults to {}.

    Returns:
        ~matplotlib.figure.Figure: Figure
        
    """

    # Load default configuration
    settings  = _default_settings(settings)

    if z is not None:
        wave = wave/(1+z)

    ferr = np.zeros_like(wave) if ferr is None else np.array(ferr)

    if mask is None:
        mask = np.zeros_like(wave, dtype=bool)

    if index_done is None:
        index_done = np.ones_like(index_names, dtype=bool)

    fig       = spectrum_simple(wave, flux, ferr, mask, ax, settings=settings)
    ax        = fig.gca()
    render    = fig.canvas.get_renderer()
    ax_heigth = np.diff(ax.get_ylim())[0]
    y_text    = ax.get_ylim()[0] + 0.05*ax_heigth
    spec_cols = [mpl.colors.to_rgb(col) for col in settings['spec_colors']]

    # Set xlim in advance to avoid problems with get_window_extent() leater
    xlim   = [min(wave[~mask][0], index_regions[0][1][0]),
              max(wave[~mask][-1], index_regions[-1][1][1])] if settings['inspect'] else \
             [min(wave[~mask][0], index_regions[index_done][0][1][0]),
              max(wave[~mask][-1], index_regions[index_done][-1][1][1]) ]  
    xspan  = xlim[1]-xlim[0]
    ax.set_xlim(xlim[0]-0.03*xspan, xlim[1]+0.03*xspan)

    # Avoid overlap between indices
    y_offset = np.zeros_like(index_names, dtype=float)
    
    for i in range(len(index_regions)):
        if (index_units[i] == 'A' or index_units[i] == 'mag'):

            # Extracting the central regions
            l, r = index_regions[i][1][0],index_regions[i][1][1]

            # Avoid overlap - to be further tested 
            if i>0:
                if 0.5*(l+r)<last_right_edge:
                    y_offset[i] += 1.2*last_texth
                
            if index_done[i]:
                t = ax_addindex(ax, text=r"$\mathrm{{{:s}}}$".format(index_names[i]), style=settings['ind_style'], 
                                bbox_index=(l,r,ax_heigth/2, ax_heigth/2), y_text=y_text+y_offset[i], 
                                color=spec_cols[2], y_text_va="bottom", fontsize=settings['spec_fontsizes'][2])
            elif settings['inspect']:
                t = ax_addindex(ax, text=r"$\mathrm{{{:s}}}$".format(index_names[i]), style=settings['ind_style'], 
                                bbox_index=(l,r,ax_heigth/2, ax_heigth/2), y_text=y_text+y_offset[i], 
                                color='darkred', y_text_va="bottom", fontsize=settings['spec_fontsizes'][2])

            _tbbox                 = t.get_window_extent(render).transformed(ax.transData.inverted()).get_points()
            last_textw, last_texth = _tbbox[1,0]-_tbbox[0,0], _tbbox[1,1]-_tbbox[0,1]
            last_right_edge        = 0.5*(l+r)+last_textw/3   # !!! empirical - to test
            
    # fig.subplots_adjust(left=0.125,bottom=0.135,right=0.95,top=0.92)

    if settings['filename']:
        fig.savefig(settings['filename'])
 
    return fig



import numpy as np

from pylick.__config__ import __table__

__all__ = ["IndexLibrary"]
    
class IndexLibrary(object):
    """Loads the index library.

    To show the full table of the available indices and assosiated keys use

    >>> from pylick.indices import IndexLibrary
    >>> lib = IndexLibrary()
    >>> print(lib)
        

    Args:
        index_keys (list, optional): identifiers of the indices to be loaded. First column of `file_table`. Defaults to None (all the available indices).
        file_table (str, optional): path to the file containing the indices definitions. Must be formatted as the default one located at `data/index_table.dat`.

    Returns: 
        IndexLibrary: Itself, to allow chaining calls.
    """

    def __init__(self, index_keys=None, file_table=None):

        if file_table is None:
            file_table = __table__
        self.source = file_table
        self._data  = self._read_index_list(file_table)

        if index_keys is not None:
            self._data  = self.__getitem__(index_keys)

        self.regions    = np.array([[v[j] for j in ['blue', 'centr', 'red']] for k, v in self._data.items()])
        self.names      = np.array([v[j] for j in ['name'] for k, v in self._data.items()])
        self.units      = np.array([v[j] for j in ['unit'] for k, v in self._data.items()])
        self.tex_names  = np.array([v[j] for j in ['tex_name'] for k, v in self._data.items()])
        self.tex_units  = np.array([v[j] for j in ['tex_unit'] for k, v in self._data.items()])


    def find(self, name):
        """Finds indices in the table.

        Args:
            name (str): the name (or part of it) of the index to search in the table.

        Returns:
            list: list of the indices found.
        """
        
        r = []
        for item in self._data.items():
            tabname = item[1]['name']
            if name in tabname:
                r.append(tabname)
        return r

    def add(self, center, blue, red, name, unit, tex_name, tex_unit):
        """.. note:: TO BE IMPLEMENTED
        Adds an index at the end of the file. Re-sorts(?) and saves(?) the library."""
        print("To be implemented")

    def pop(self, key):
        """.. note:: TO BE IMPLEMENTED
        Removes an index from the library. Re-sorts(?) the library."""
        print("To be implemented")

    @classmethod
    def _read_index_list(self, file_table):
        """Read index table

        Args:
            file_table (str): path to the file containing the indices definitions.

        Returns:
            dict: dictionary of indices and associated spectral regions (name, unit, centr, blue, red, tex_name, tex_unit).
        """

        data = {}

        with open(file_table, 'r') as f:
            data = {}
            for i, line in enumerate(f):
                if line[0] != '#':
                    l = line.split()
                    try:
                        attr = dict(
                            centr       = (float(l[3]), float(l[4])),
                            blue        = (float(l[5]), float(l[6])),
                            red         = (float(l[7]), float(l[8])),
                            name        = l[1],
                            unit        = l[2],
                            tex_name    = l[9],
                            tex_unit    = l[10]
                        )
                    except IndexError:
                        print("> Error while loading the index table at line: {:d}. Perhaps a `#` is missing?".format(i))
                    except:
                        print("> Error while loading the index table. It must be formatted as the default one located at `data/index_table.dat`.")
                    ID = int(l[0])
                    data[ID] = attr
        return data

    def __repr__(self):
        return "IndexLibrary(file_table='{:s}')".format(self.source)


    def __str__(self):
        msg = "{:4s}{:9s}{:10s}{:19s}{:19s}{:19s}{:10s}\n".format(
            "ID", "name", "unit", "blue", "centr", "red", "tex_name")
        msg = msg + '-'*len(msg) + '\n'
        for key, val in self._data.items():
            msg = msg + "{:<4d}{:9s}{:10s}{:8.3f}-{:8.3f}  {:8.3f}-{:8.3f}  {:8.3f}-{:8.3f}  {:s}\n".format(
                key,val['name'],val['unit'],val['blue'][0],val['blue'][1],val['centr'][0],val['centr'][1],
                val['red'][0],val['red'][1],val['tex_name'])
        return msg  


    def __len__(self):
        return len(self._data)


    def __getitem__(self, keys):
        d = self._data
        return {k:d[k] for k in keys if k in d}



import os
import numpy as np
from astropy.io import fits



def _dir_build(path, dir_name, force=False):
    r""" Checks and builds `dir_name`

    Args:
        path (str): path of the folder to build
        dir_name (char): name of the folder to build
        force (bool): remove folder if already existing. Defaults to False.
    """

    full_path = path+'/'+dir_name+'/'

    if (not os.path.exists(full_path)):
        print("\nFolder {:s} created\n".format(dir_name))
        os.mkdir(full_path)

    elif force:
        print("\nFolder {:s} & its content are removed\n".format(dir_name))
        os.system('rm -r ' + full_path)
        os.mkdir(full_path)   

    else:
        print("\nWARNING: The folder {} already exists, set 'force=True' to remove it\n".format(full_path))



def _detect_spectrum_window(flux):
    r"""Finds the spectrum window excluding non-positive side regions

    Args:
        flux (np.ndarray): spectral fluxes

    Returns:
        np.ndarray: Boolean mask to remove side regions where flux is non-posivive
    """
    il   = 0
    ir   = -1
    while True:
        if (flux[il] > 0):
            break
        il += 1
    while True:
        if (flux[ir] > 0):
            break
        ir -= 1        
    flag_window = np.ones_like(flux, dtype=bool)
    flag_window[:il] = 0
    flag_window[ir:] = 0

    return flag_window



def load_spec_fits(dir_spec, filename, colnames=['wave', 'flux', 'ferr', 'mask'], reduce_window=False, hdul=1):
    r""" Function to load a .fits spectrum

    Args:
        dir_spec (str): directory where the spectrum is located
        filename (str): filename including the .fits extension
        colnames (list, optional): Names of the fits columns. Defaults to ['wave', 'flux', 'ferr', 'mask'].
        reduce_window (bool, optional): Whether to restrict the spectrum excluding non-positive side regions. Defaults to False.
        hdul (int, optional): N of the extension of the .fits Header Data Unit where the spectrum is located. Defaults to 1.

    Returns:
        list: spectrum imported according to the given colnames
    """
    spectrum = [[]]*len(colnames)

    hdulist  = fits.open(dir_spec+filename)

    if reduce_window:
        window      = _detect_spectrum_window(hdulist[hdul].data[colnames[hdul]][0])

        try:
            spectrum    = [hdulist[hdul].data[colnames[i]][0][window] for i in range(len(colnames))]
        except KeyError as e:
            print("> PyLick: {:s} - These are the available keys: {:s}".format(e, hdulist[hdul].columns))
    
    else:
        try:
            spectrum    = [hdulist[hdul].data[colnames[i]][0] for i in range(len(colnames))]
        except KeyError as e:
            print("> PyLick: {:s} - These are the available keys: {:s}".format(e, hdulist[hdul].columns))
    
    hdulist.close()

    return np.array(spectrum)



def spec_stats(wave, flux, ferr=None):
    r""" Prints statistics of the spectrum

    Args:
        wave (np.ndarray): wavelength sampling of the spectrum
        flux (np.ndarray): spectral fluxes corresponding to `spec_wave`
        ferr (np.ndarray, optional): spectral fluxes' uncertainties.  Defaults to None.
    """

    pars = ["wave", "flux"] if ferr is None else ["wave", "flux", "ferr", " SNR"]
    vals = [wave, flux] if ferr is None else [wave, flux, ferr]
    
    for p, v in zip(pars, vals):
        print(" {:s})\t range: {}--{}\t mean: {}\t median: {}" \
              .format(p, *np.around([v.min(), v.max(), np.mean(v), np.median(v)], decimals=2),)
             )

    print(" delta_wave_mean: {:.2f}".format(np.mean(np.diff(wave))))




import time
import warnings
# warnings.filterwarnings("ignore")
import numpy as np

from astropy.table import Table, MaskedColumn

from pylick.plot import spectrum_with_indices
from pylick.indices import IndexLibrary
from pylick.measurement import MeasSpectrum
from pylick.__config__ import __table__, __dirRes__


__all__ = ['Catalog', 'Galaxy']


def printConsole(text, color):
    if color == 'w':
        print("\x1b[0;49;97m"+text+"\x1b[0m")
    if color == 'c':
        print("\x1b[36m"+text+"\x1b[0m")
    if color == 'r':
        print("\x1b[91m"+text+"\x1b[0m")
    if color == 'y':
        print("\x1b[1;33m"+text+"\x1b[0m")

def helloMsg():
    printConsole("################################################################", 'r')
    printConsole("#### pyLick: A Python program to measure spectral indices.  ####", 'r')
    printConsole("####                                                        ####", 'r')
    printConsole("####        Authors: M. Moresco, N. Borghi, S. Quai         ####", 'r')
    printConsole("################################################################\n", 'r')

def progressBar(iteration, total, prefix = 'Progress:', suffix = '', length = 40, fill = '█'):
    prefix =  ("{:8s}").format(prefix)
    percent = ("{:d}").format(int(100 * (iteration / float(total))))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    # print('\rID {:s}: {:s}\t'.format(prefix, bar), end = "")
    print(f'  ID {prefix}: |{bar}| {percent}% {suffix}', end = "\r")

def printSingleRes(ID, names, vals, errs, BPR=None):
    printConsole("\n < Galaxy ID {:s} >".format(ID),'y')
    ind_table                   = Table(masked=True)
    ind_table["Index"]          = names
    ind_table["Value"]          = MaskedColumn(data=vals, mask=np.isnan(vals))
    ind_table["Error"]          = MaskedColumn(data=errs, mask=np.logical_or(np.isnan(vals), errs==0) )
    ind_table["BPR (%)"]        = BPR*100
    ind_table["Value"].format   = "{:.4f}"
    ind_table["Error"].format   = "{:.4f}"
    ind_table["BPR (%)"].format = "{:.1f}"
    ind_table.remove_rows(np.where(np.isnan(vals))[-1]) # Remove if index is not measured
    ind_table.pprint_all()

class Catalog:
    """Class optimized to handle the analysis of an entire catalog of galaxies. 

    >>> from pylick.analysis import Catalog
    >>> meas_table = Catalog(...)
    >>> print(meas_table)

    Args:
        IDs (str): galaxy unique identifier(s)
        load_spectrum (funct): takes one `ID` as argument and returns the corresponding spectral [wave, flux, err, mask].
        index_keys (list): keys of the indices to measure (see :class:`pylick.indices.IndexLibrary`).
        meas_method (str, optional): method to measure the indices ("int", "wei", "exact"). Defaults to "int".
        z (list, optional): if given, `wave`s are assumed to be in the obs. frame and shifted to rest-frame. Defaults to None.
        bpr_thres (float, optional): bad-to-total pixel ratio above which the measurement is not performed. Defaults to 1.
        file_table (str, optional): path to the file containing the indices definitions. Must be formatted as the default one located at `data/file_table.dat`.
        plot (bool, optional): produce plot file. Defaults to False.
        plot_settings (dict, optional): instruction for the plot code. Defaults to {}.
        verbose (bool, optional): prints warnings and execution time in the console. Defaults to False.
    """        


    def __init__(self, IDs, load_spectrum, index_keys, 
                 z=None, meas_method='int', bpr_thres=1., file_table=None, 
                 plot=False, plot_settings={}, verbose=False):    

        self.IDs            = np.array(IDs, dtype=str)
        self.index_keys     = index_keys
        self.meas_method    = meas_method
        self.bpr_thres      = bpr_thres
        self.file_table     = file_table 
        self.Nobj           = len(IDs)
        self.Nidx           = len(index_keys)
        self.file_table     = __table__ if self.file_table is None else file_table
        lib                 = IndexLibrary(file_table=file_table, index_keys=index_keys)

        data_names          = ['']*self.Nidx*2
        data_names[::2]     = lib.names
        data_names[1::2]    = [name+'_err' for name in lib.names]
        data_indices        = np.empty((self.Nobj, 2*self.Nidx), dtype=float)

        # GO
        t0 = time.time()
        helloMsg()

        # if not verbose: progressBar(0, self.Nobj)
        for i, ID in enumerate(self.IDs):

            if not verbose: progressBar(i, self.Nobj, ID)

            spectrum   = np.array(load_spectrum(ID))
            wave       = spectrum[0] if z is None else spectrum[0]/(1+z[i])
            flux       = spectrum[1]
            ferr       = spectrum[2]
            mask       = spectrum[3].astype(bool)
            ferr[mask] = 9.9*10**99

            assert wave.size == flux.size == ferr.size, \
                    "pyLick: wave, flux, (ferr) must have the same size"
            
            meas   = MeasSpectrum(wave=wave, flux=flux, ferr=ferr, todo=~mask,
                                  regions=lib.regions, units=lib.units, names=lib.names, 
                                  meas_method=self.meas_method, BPR_thres=self.bpr_thres, 
                                  nans=np.nan, verbose=False)

            names  = lib.names
            vals   = meas.vals 
            errs   = meas.errs

            # Fill results array ind1, ind1_err, ind2, ind2_err, ...
            data_indices[i][::2]  = vals
            data_indices[i][1::2] = errs

            if plot:
                spectrum_with_indices(ID, wave, flux, ferr, mask,
                                      lib.regions, lib.tex_names, lib.units, meas.finite,
                                      plot_settings)

            if verbose: printSingleRes(ID, names, vals, errs, meas.BPR)

        tf = time.time() - t0

        self.data_names   = data_names
        self.data_indices = data_indices

        if not verbose: 
            progressBar(i+1, self.Nobj, ID, suffix=" in {:.1f}s ({:.2f}/s)".format(tf, tf/self.Nobj))
            print()                
        else:
            print()
            printConsole("Elapsed time: {:.2f} s".format(tf), 'r')
            print()


    @property
    def results(self):
        """Retrieve the results table."""
        t =  Table(self.data_indices, names=self.data_names)
        t.add_column(self.IDs, name="ID", index=0)
        return t

    def __repr__(self):
        return "Catalog Class - use `.data_res` attributes to retrieve the measurements astropy.Table, alternatively print()"

    def __str__(self):
        return self.results

    def save(self, filename, filefmt="ascii.ecsv", prepend_colnames="", prepend_table=None):
        """Save the results table.

        Args:
            filename (str): filename.
            filefmt (str, optional): format of the :class:`astropy.Table` to save. Defaults to "ascii.ecsv".
            prepend_colnames (str, optional): string to which append index names to be used as catalog columns. Defaults to empty string.
            prepend_table (:class:`astropy.Table`, optional): catalog to which append the results. Defaults to None.
        """

        out_colnames = [prepend_colnames+x for x in self.data_names]

        if prepend_table is None:
            res_table       = Table(masked=True)
            res_table["ID"] = self.IDs
        else:
            res_table = Table(prepend_table, masked=True)

        for i, colname in enumerate(out_colnames):
            res_table.add_column(self.data_indices[:,i], name=colname)
        
        res_table.write(filename, overwrite=True)

    # end Catalog class



class Galaxy:
    """Class optimized to handle the analysis of a single galaxy. 

    >>> from pylick.analysis import Galaxy
    >>> ind_meas = Galaxy(...)
    >>> print(ind_meas)

    Args:
        ID (str): galaxy unique identifier
        index_keys (list): keys of the indices to measure (see :class:`pylick.indices.IndexLibrary`).
        wave (np.ndarray): wavelength sampling of the spectrum.
        flux (np.ndarray): spectral fluxes corresponding to `wave`.
        ferr (np.ndarray, optional): spectral fluxes uncertainties. Defaults to None.
        mask (np.ndarray, optional): boolean array flagging bad (True) and good (False) pixels. Defaults to None.
        z (float, optional): if given, `wave` is assumed to be in the obs. frame and shifted to rest-frame. Defaults to None.
        meas_method (str, optional): method to measure the indices ("int", "wei", "exact"). Defaults to "int".
        bpr_thres (float, optional): bad-to-total pixel ratio above which the measurement is not performed. Defaults to 1.
        file_table (str, optional): path to the file containing the indices definitions. Must be formatted as the default one located at `data/file_table.dat`.
        verbose (bool, optional): prints warnings and execution time in the console. Defaults to True.
    """        

    def __init__(self, ID, index_keys, wave, flux,
                 ferr=None, mask=None, z=None,
                 meas_method='int', bpr_thres=1., file_table=None,
                 plot=False, plot_settings={}, verbose=True):

        self.IDs           = ID
        self.index_keys    = index_keys
        self.wave          = np.array(wave)
        self.flux          = np.array(flux)
        self.meas_method   = meas_method
        self.bpr_thres     = bpr_thres
        self.file_table    = file_table 
        self.Nidx          = len(index_keys)

        self.file_table = __table__ if self.file_table is None else file_table

        self.wave = self.wave if z is None \
                    else self.wave/(1+z)

        self.ferr = np.zeros_like(self.wave) if ferr is None \
                    else np.array(ferr)
        
        self.mask = np.zeros_like(self.wave, dtype=bool) if mask is None \
                    else np.array(mask, dtype=bool) 

        assert self.wave.size == self.flux.size == self.ferr.size, \
                "pyLick: wave, flux, (ferr) must have the same size"

        # GO
        if verbose: t0 = time.time()

        self.qual             = (~self.mask)   # pixels to keep
        self.ferr[~self.qual] = 9.9*10**99.

        self.lib              = IndexLibrary(file_table=self.file_table, index_keys=self.index_keys)

        self.meas             = MeasSpectrum(self.wave, self.flux, self.ferr, self.qual,
                                             self.lib.regions, self.lib.units, self.lib.names, self.meas_method,
                                             BPR_thres=self.bpr_thres, nans=np.nan, verbose=verbose)


        if verbose: printConsole("Elapsed time: {:.2f} s\n".format(time.time() - t0), 'r')


    @property
    def names(self):
        """Retrieve indices names."""
        return self.lib.names

    @property
    def vals(self):
        """Retrieve indices values."""
        return self.meas.vals

    @property
    def errs(self):
        """Retrieve indices uncertainties."""
        return self.meas.errs

    def __repr__(self):
        return "Galaxy Class - use `.names`, `.vals`, `.errs` attributes to retrieve the measurements, alternatively print()"

    def __str__(self):
        msg = "{:10s}      {:5s}      {:5s}\n".format("names", "vals", "errs")
        msg = msg + '-'*len(msg) + '\n'

        for i in range(self.Nidx):
            if np.isnan(self.vals[i]) & np.isnan(self.errs[i]):
                continue
            else:
                msg = msg + "{:10s} {:10.4f} {:10.4f}\n".format(self.names[i], self.vals[i], self.errs[i])
        return msg  

    def plot(self, filename=None, settings={}):
        """ Produce the plot with annotated indices 
            (see :py:func:`pylick.plot.spectrum_with_indices`).

            Args:
                plot (bool, optional): produce plot file. Defaults to False.
                plot_settings (dict, optional): instructions for the plot code. Defaults to {}.
            
            Returns:
                ~matplotlib.figure.Figure: Figure
        """

        fig = spectrum_with_indices(self.wave, self.flux, self.ferr, mask=self.mask,
                                    index_regions=self.lib.regions, index_names=self.lib.tex_names, 
                                    index_units=self.lib.units, index_done=self.meas.finite,
                                    ax=None, settings=settings)

        if filename is None:
            fig.show()
        else:
            fig.savefig(filename)
        

    # end Galaxy class



import numpy as np
from scipy import interpolate, integrate


__all__ = [
    'MeasSpectrum'
    ]


def _direct_integration(y,x):
    res=0
    for i in range(len(x)-1):
        res+=y[i]*(x[i+1]-x[i])
    return res



def _wider_slice(todo, width=1):
    ''' Enlarge a slice of \pm width

    e.g. width=1,  0 0 0 1 1 0 0 -> 0 0 1 1 1 1 0

    '''
    tmptodo  = np.array(todo, dtype=bool)

    for i in range(width, np.size(tmptodo)-width+1):
        if todo[i] == True:
            tmptodo[i-width:i+width+1] = True

    return tmptodo



class MeasSpectrum:
    """Measurement of multiple indices on a single spectrum. 

    Args:
        IDs (str): galaxy unique identifier(s)
        index_list (list): list of indices to measure
        wave (np.ndarray): wavelength sampling of the spectrum
        flux (np.ndarray): spectral fluxes corresponding to `wave`
        ferr (np.ndarray, optional): spectral fluxes uncertainties.  Defaults to None.
        todo (np.ndarray, optional): boolean array flagging pixels to interpolate over. Defaults to None.
        z (float, optional): if given, `wave` are computed rest-frame. Defaults to None.
        bpr_thres (float, optional): bad-to-total pixel ratio above which the measurement is not performed. Defaults to None.
        meas_method (str, optional): method to measure the indices. Defaults to 'int'.
        index_table (str, optional): path to the user-defined table of indices. Defaults to None.
        plot (bool, optional): produce plot file. Defaults to False.
        plot_settings (dict, optional): instruction for the plot code. Defaults to {}.
    """

    def _slice_spec(self, wave, flux, ferr, todo, l, r, enlarge=None):
        """Creates slices of spectra given left and right margins

        Args:
            l (float): left margin
            r (float): right margin
            enlarge (int, optional): number of pixel to enlarge the region. Defaults to None.

        Returns:
            list: wave, flux, ferr arrays, number of total and bad pixels
        """
        clip = (wave > l) & (wave < r)

        if enlarge is not None:
            clip = _wider_slice(clip, width=enlarge)

        slice_wave  = wave[clip]
        slice_flux  = flux[clip]
        slice_ferr  = ferr[clip]
        n_pix_total = np.sum(clip)
        n_pix_bad   = np.sum(~todo[clip])
        
        # if 0:
        #     print("Slices of wave, flux, ferr")
        #     print(slice_wave)
        #     print(slice_flux)
        #     print(slice_ferr)
            # print("Total pixels: {}, of which bad {}".format(n_pix_total, n_pix_bad))

        return slice_wave, slice_flux, slice_ferr, n_pix_total, n_pix_bad


    def _flux_interp_np(self, l, r, binsize):
        """Interpolates the spetrum around the region [l,r] -/+ binsize

        Args:
            l (float): left margin
            r (float): right margin
            binsize ([type]): pixel size of the new interpolated array
        """        
        self.wave_intp = np.arange(l-binsize, r+binsize, binsize)
        self.flux_intp = np.interp(self.wave_intp, self.wave[self.todo],
                                        self.flux[self.todo])
        self.ferr_intp  = np.sqrt( np.interp(self.wave_intp, self.wave[self.todo], 
                                                np.square(self.ferr[self.todo])) )
        self.todo_intp = np.ceil( np.interp(self.wave_intp, self.wave, self.todo) ).astype(bool)


    def _flux_interp_exact(self, l, r, binsize):
        """Generates the interpolation function around the region [l,r] -/+ binsize

        Args:
            l (float): left margin
            r (float): right margin
            binsize ([type]): pixel size of the new interpolated array
        """    

        todo_intp = self.todo[self.todo]

        self.wave_intp = np.arange(l-binsize, r+binsize, binsize)
        self.flux_intp = interpolate.interp1d(self.wave[self.todo],self.flux[self.todo], kind=self.intp_kind)
        self.ferr_intp = interpolate.interp1d(self.wave[self.todo],self.ferr[self.todo], kind=self.intp_kind)

        # self.ferr_intp  = np.sqrt( np.interp(self.wave_intp, self.wave[self.todo], 
        #                                         np.square(self.ferr[self.todo])) )


    def _check_ind_def(self, regions, unit):
        msg = "PyLick: index regions are not properly defined"

        bb, br, cb, cr, rb, rr = regions.flatten()    

        if (unit == 'A') or (unit == 'mag'):
            assert bb < br and cb < cr and rb < rr, msg

    # ---------------------------------------------------------
    def _measure_multi_exact(self, regions, unit):
        """

        Parameters
        __________
        
        x : np.array
            spectrum wavelenghts

        y: np.array
            spectrum flux

        yerr : np.array

        
        regions : [np.float]*6
            index definition: blue pseudocont. [bb,br], central [cb, cr], red pseudocont. [rb, rr] 

        """



        bb, br, cb, cr, rb, rr = regions.flatten()              # Index definition bands
        bc, cc, rc = 0.5*(bb+br), 0.5*(cb+cr), 0.5*(rb+rr)		# Mean points
        db, dc, dr = (br - bb), (cr - cb), (rr - rb)			# Widths



        ind_value = self.nan
        ind_error = self.nan
        ind_BPR   = -1

        min_spec = np.min(self.wave[self.todo])
        max_spec = np.max(self.wave[self.todo])

        if (bb>min_spec) and (rr<max_spec): 

            # Interpolate the spectrum around the index region
            # self._flux_interp_exact(bb,rr,binsize=self.int_binsize)

            w_blu, f_blu, n_blu, Nblu, Nbad_blu = self._slice_spec(wave=self.wave,
                                                                   flux=self.flux,
                                                                   ferr=self.ferr,
                                                                   todo=self.todo,
                                                                   l=bb, r=br)
            w_cen, f_cen, n_cen, Ncen, Nbad_cen = self._slice_spec(wave=self.wave,
                                                                   flux=self.flux,
                                                                   ferr=self.ferr,
                                                                   todo=self.todo,
                                                                   l=cb, r=cr)
            w_red, f_red, n_red, Nred, Nbad_red = self._slice_spec(wave=self.wave,
                                                                   flux=self.flux,
                                                                   ferr=self.ferr,
                                                                   todo=self.todo,
                                                                   l=rb, r=rr)
            
            # flux_blu_mean = self.flux_intp.integral(bb, br)/(br-bb)
            # flux_red_mean = self.flux_intp.integral(rb, rr)/(rr-rb)
            flux_blu_mean = integrate.quad(self.flux_intp, bb, br, points=w_blu, limit=Nblu+2)[0] / (br-bb)
            flux_red_mean = integrate.quad(self.flux_intp, rb, rr, points=w_red, limit=Nred+2)[0] / (rr-rb)
            
            if (unit == 'A') or (unit == 'mag'):
                
                assert bb < br and cb < cr and rb < rr, "PyLick: index regions are not properly defined"

                # Pseudocontinuum estimated by linearly interpolating mean fluxes of lateral regions
                flux_pseudoc  = lambda x : ( flux_blu_mean * (rc - x) + flux_red_mean * (x - bc) ) / (rc - bc)

                ##### Errors from Cardiel et al. 1998
                # Number of pixels in the three bandpass
                Ntot	= Nblu + Ncen + Nred
                # OCIO!!!! To avoid errors from noise=0
                FNR     = np.sum(np.divide(f_blu,n_blu)) + np.sum(np.divide(f_cen,n_cen)) + np.sum(np.divide(f_red,n_red))
                SNR     = ( 1./(Ntot*np.sqrt(self.disp)) ) * FNR	# SN[AA]

                c2 = np.sqrt((1./dc)+
                                (((rc-cc)/(rc-bc))**2)*(1./db) +
                                (((cc-bc)/(rc-bc))**2)*(1./dr))
                c1 = dc*c2				
                c3 = 2.5*c2*np.log10(np.e)

                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_cen+Nbad_red)/Ntot

                if unit == 'A':
                    fcn 		= lambda x : (1 - self.flux_intp(x) / flux_pseudoc(x))
                    ind_value	= integrate.quad(fcn, cb, cr, points=w_cen,  limit=Ncen+2)
                    # print(ind_value)
                    ind_value = ind_value[0]
                    ind_error 	= (c1-(c2*ind_value)) / SNR

                if unit == 'mag':
                    fcn 		= lambda x : (self.flux_intp(x) / flux_pseudoc(x))
                    ind_value	= -2.5 * np.log10( integrate.quad(fcn, cb, cr, points=w_cen, limit=Ncen+2)[0] / (cr - cb) )
                    ind_error 	= c3 / SNR

            elif unit == 'break_nu':
                flux_nu		  = lambda x : x**2 * self.flux_intp(x)
                noise_nu   	  = lambda x : x**2 * self.ferr_intp(x)
                flux_blu_mean = integrate.quad(flux_nu, bb, br, points=w_blu, limit=Nblu+2)[0] / (br-bb)
                flux_red_mean = integrate.quad(flux_nu, rb, rr, points=w_red, limit=Nred+2)[0] / (rr-rb)
                ind_value	  = flux_red_mean/flux_blu_mean

                #ERROR
                SNR_red		= (1./(Nred*np.sqrt(self.disp))) * (np.sum( flux_nu(w_red)/noise_nu(w_red) ))
                SNR_blue	= (1./(Nblu*np.sqrt(self.disp))) * (np.sum( flux_nu(w_blu)/noise_nu(w_blu) ))
                ind_error 	= ind_value * np.sqrt((1./np.square(SNR_blue)) + (1./np.square(SNR_red))) / np.sqrt(dr)
                
            
                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_red) / (Nblu+Nred) 

            elif unit == 'break_lb':
                ind_value	= flux_red_mean/flux_blu_mean

                #ERROR
                SNR_red		= (1./(Nred*np.sqrt(self.disp))) * (np.sum(np.divide(f_red,n_red)))
                SNR_blue	= (1./(Nblu*np.sqrt(self.disp))) * (np.sum(np.divide(f_red,n_red)))
                ind_error 	= ind_value * np.sqrt((1./np.square(SNR_blue)) + (1./np.square(SNR_red))) / np.sqrt(dr) #ocio

                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_red) / (Nblu+Nred) 

            else:
                print(unit)

        return ind_value, ind_error, ind_BPR


    def _measure_multi_interp(self, regions, unit):
        """ Interpolation method

        Args:
            regions (float): [[bb,br], [cb,cr], [rb, rr]] i.e. [blue pseudocont., central, red pseudocont.] 
            unit (str): unit type (A, mag, break_nu, break_lb), this defines the index

        Returns:
            ndarray: ind_value, ind_error, ind_BPR
        """

        bb, br, cb, cr, rb, rr = regions.flatten()              # Index definition bands

        ind_value = self.nan
        ind_error = self.nan
        ind_BPR   = -1

        min_spec = np.min(self.wave[self.todo])
        max_spec = np.max(self.wave[self.todo])

        
        if (bb>min_spec) and (rr<max_spec): 
            
            # Interpolate the spectrum around the index region
            self._flux_interp_np(bb,rr,binsize=self.int_binsize)


            bc, cc, rc = 0.5*(bb+br), 0.5*(cb+cr), 0.5*(rb+rr)		# Mean points
            db, dc, dr = (br - bb), (cr - cb), (rr - rb)			# Widths

            
            w_blu, f_blu, n_blu, Nblu, Nbad_blu = self._slice_spec(wave=self.wave_intp, 
                                                                   flux=self.flux_intp,
                                                                   ferr=self.ferr_intp,
                                                                   todo=self.todo_intp,
                                                                   l=bb,r=br)
            w_cen, f_cen, n_cen, Ncen, Nbad_cen = self._slice_spec(self.wave_intp, 
                                                                   self.flux_intp,
                                                                   self.ferr_intp,
                                                                   self.todo_intp,
                                                                   l=cb, r=cr)
            w_red, f_red, n_red, Nred, Nbad_red = self._slice_spec(self.wave_intp, 
                                                                   self.flux_intp,
                                                                   self.ferr_intp,
                                                                   self.todo_intp,
                                                                   l=rb, r=rr)
            
            flux_blu_mean = f_blu.mean()
            flux_red_mean = f_red.mean()
            wave_blu_mean = w_blu.mean()
            wave_red_mean = w_red.mean()

            if (unit == 'A') or (unit == 'mag'):
                
                assert bb < br and cb < cr and rb < rr, "PyLick: index regions are not properly defined"

                # Estimate pseudo-continuum
                m           = (flux_red_mean-flux_blu_mean)/(wave_red_mean-wave_blu_mean)
                q           = flux_blu_mean - m*wave_blu_mean
                f_pseudo    = m * (w_cen) + q

                ##### ERROR (Cardiel+1998)
                # Number of pixels in the three bandpass
                Ntot	= Nblu + Ncen + Nred
                # Flux to noise ratio
                FNR     = np.sum(np.divide(f_blu,n_blu)) + np.sum(np.divide(f_cen,n_cen)) + np.sum(np.divide(f_red,n_red))
                SNR     = ( 1./(Ntot*np.sqrt(self.disp)) ) * FNR	# SN[AA]
                c2 = np.sqrt((1./dc)+
                                (((rc-cc)/(rc-bc))**2)*(1./db) +
                                (((cc-bc)/(rc-bc))**2)*(1./dr))
                c1 = dc*c2				
                c3 = 2.5*c2*np.log10(np.e)

                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_cen+Nbad_red) / Ntot 

                if unit == 'A':
                    ind_value = integrate.simps(1-(f_cen/f_pseudo),w_cen,even='avg')
                    ind_error = (c1-(c2*ind_value)) / SNR

                elif unit == 'mag':
                    ind_value = -2.5*np.log10(integrate.simps(f_cen/f_pseudo,w_cen,even='avg')/dc)
                    ind_error = c3 / SNR


            elif unit == 'break_nu':
                flux_nu_red  = np.multiply(np.square(w_red),f_red)
                flux_nu_blu  = np.multiply(np.square(w_blu),f_blu)
                noise_nu_red = np.multiply(np.square(w_red),n_red)
                noise_nu_blu = np.multiply(np.square(w_blu),n_blu)
                SNR_red		= (1./(Nred*np.sqrt(self.disp))) * (np.sum(np.divide(flux_nu_red,noise_nu_red)))
                SNR_blue	= (1./(Nblu*np.sqrt(self.disp))) * (np.sum(np.divide(flux_nu_blu,noise_nu_blu)))

                ind_value   = flux_nu_red.mean()/flux_nu_blu.mean()
                ind_error 	= ind_value * np.sqrt((1./np.square(SNR_blue)) + (1./np.square(SNR_red))) / np.sqrt(dr)

                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_red) / (Nblu+Nred) 

            elif unit == 'break_lb':
                SNR_red		= (1./(Nred*np.sqrt(self.disp))) * (np.sum(np.divide(f_red,n_red)))
                SNR_blue	= (1./(Nblu*np.sqrt(self.disp))) * (np.sum(np.divide(f_blu,n_blu)))
                
                ind_value	= flux_red_mean/flux_blu_mean
                ind_error 	= ind_value * np.sqrt((1./np.square(SNR_blue)) + (1./np.square(SNR_red))) / np.sqrt(dr) #ocio

                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_red) / (Nblu+Nred) 

            else:
                print(unit)

        else:
            if self.debug:
                print("pylick: Index outside of the spectrum ({:.2f} < lambda < {:.2f})".format(min_spec, max_spec))


        return ind_value, ind_error, ind_BPR


    def _measure_multi_wei(self, regions, unit):
        
        # Index definition bands = [bb, br], [cb, cr], [rb, rr]
        bb, br, cb, cr, rb, rr = regions.flatten()              

        ind_value = self.nan
        ind_error = self.nan
        ind_BPR   = -1

        if (bb>self.wave.min()) and (rr<self.wave.max()):

            bc, cc, rc = 0.5*(bb+br), 0.5*(cb+cr), 0.5*(rb+rr)		# Mean points
            db, dc, dr = (br - bb), (cr - cb), (rr - rb)			# Widths

            w_blu, f_blu, n_blu, Nblu, Nbad_blu = self._slice_spec(bb, br, enlarge=1)
            w_cen, f_cen, n_cen, Ncen, Nbad_cen = self._slice_spec(cb, cr, enlarge=1)
            w_red, f_red, n_red, Nred, Nbad_red = self._slice_spec(rb, rr, enlarge=1)
            
            weight_b              = np.ones(Nblu, dtype=float)
            weight_r              = np.ones(Nred, dtype=float)
            weight_b[0]           = (w_blu[1] - bb) / (w_blu[1] - w_blu[0])
            weight_r[0]           = (w_red[1] - rb) / (w_red[1] - w_red[0])
            weight_b[-1]          = (br - w_blu[-2]) / (w_blu[-1] - w_blu[-2])
            weight_r[-1]          = (rr - w_red[-2]) / (w_red[-1] - w_red[-2])
            weight_b[weight_b>1.] = 1.
            weight_r[weight_r>1.] = 1.

            flux_blu_mean = np.average(f_blu, weights=weight_b)
            flux_red_mean = np.average(f_red, weights=weight_r)
            
            if (unit == 'A') or (unit == 'mag'):
                weight_c              = np.ones(Ncen, dtype=float)
                weight_c[0]           = (w_cen[1] - cb) / (w_cen[1] - w_cen[0])
                weight_c[-1]          = (cr - w_cen[-2]) / (w_cen[-1] - w_cen[-2])
                weight_c[weight_c>1.] = 1.
                
                # Estimate pseudo-continuum
                bc            = np.average(w_blu, weights=weight_b)
                rc            = np.average(w_red, weights=weight_r)
                m             = (flux_red_mean-flux_blu_mean)/(rc-bc)
                q             = flux_blu_mean - m*bc
                f_pseudo      = m * (w_cen) + q

                ##### ERROR
                # Number of pixels in the three bandpass
                Ntot	= Nblu + Ncen + Nred
                # Flux to noise ratio
                FNR     = np.sum(np.divide(f_blu*weight_b,n_blu)) + \
                        np.sum(np.divide(f_cen*weight_c,n_cen)) + \
                        np.sum(np.divide(f_red,n_red))

                SNR     = ( 1./(Ntot*np.sqrt(self.disp)) ) * FNR

                c2 = np.sqrt((1./dc)+
                            (((rc-cc)/(rc-bc))**2)*(1./db) +
                            (((cc-bc)/(rc-bc))**2)*(1./dr))
                c1 = dc*c2				
                c3 = 2.5*c2*np.log10(np.e)

                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_cen +Nbad_red) / Ntot
                # ind_BPR = (Nbad_cen) / Ncen


                if (unit=='A') :
                    #ind_value=simps((1-(flux_ind/pc_ind))*weight_ind,lambda_ind,even='avg')	
                    #ind_value=trapz(1-(flux_ind/pc_ind),lambda_ind)	
                    ind_value=_direct_integration((1-(f_cen/f_pseudo))*weight_c,w_cen)
                    ind_error = (c1-(c2*ind_value)) / SNR

                elif (unit=='mag') :
                    #ind_value=-2.5*log10(simps(flux_ind/pc_ind*weight_ind,lambda_ind,even='avg')/(cr-cb))	
                    #ind_value=-2.5*log10(trapz(flux_ind/pc_ind,lambda_ind)/(cr[i]-cb[i]))	
                    ind_value=-2.5*np.log10(_direct_integration(f_cen/f_pseudo*weight_c,w_cen)/dc)
                    ind_error = c3 / SNR


            elif unit == 'break_nu':
                flux_nu_red  = np.multiply(np.square(w_red),f_red)
                flux_nu_blu  = np.multiply(np.square(w_blu),f_blu)
                noise_nu_red = np.multiply(np.square(w_red),n_red)
                noise_nu_blu = np.multiply(np.square(w_blu),n_blu)
                disp_r       = np.mean(np.diff(w_red))
                disp_b       = np.mean(np.diff(w_blu))
                SNR_red		 = (1./(Nred*np.sqrt(disp_r))) * (np.sum(np.divide(flux_nu_red*weight_r,noise_nu_red)))
                SNR_blue	 = (1./(Nblu*np.sqrt(disp_b))) * (np.sum(np.divide(flux_nu_blu*weight_b,noise_nu_blu)))

                ind_value    = np.average(flux_nu_red, weights=weight_r) / np.average(flux_nu_blu, weights=weight_b)
                ind_error 	 = ind_value * np.sqrt((1./np.square(SNR_blue)) + (1./np.square(SNR_red))) / np.sqrt(dr)
                
                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_red) / (Nblu+Nred) 

            elif unit == 'break_lb':
                disp_r       = np.mean(np.diff(w_red))
                disp_b       = np.mean(np.diff(w_blu))
                SNR_red		 = (1./(Nred*np.sqrt(disp_r))) * (np.sum(np.divide(f_red*weight_r,n_red)))
                SNR_blue	 = (1./(Nblu*np.sqrt(disp_b))) * (np.sum(np.divide(f_blu*weight_b,n_blu)))
                
                ind_value	 = flux_red_mean/flux_blu_mean
                ind_error 	 = ind_value * np.sqrt((1./np.square(SNR_blue)) + (1./np.square(SNR_red))) / np.sqrt(dr)

                # Bad Pixels Ratio
                ind_BPR = (Nbad_blu+Nbad_red) / (Nblu+Nred) 

        return ind_value, ind_error, ind_BPR


    # ------------------------------------------- To Be Done...
    def _measure_ind_general_int(self,x,y,erry,disp):
        # bb,br = blue 
        # cb,cr = central
        # rb,rr = red
        print(" This feature is not available yet.")
        if False:
            # Ref. Cenarro et al. (2001)
            # pseudocontinuum:
            flux_generic_pc=y[(x>=8474.)*(x<=8484.)+(x>=8563.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            noise_generic_pc=erry[(x>=8474.)*(x<=8484.)+(x>=8563.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]		
            lambda_generic_pc=x[(x>=8474.)*(x<=8484.)+(x>=8563.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            pc = np.polyfit(lambda_generic_pc, flux_generic_pc, 1, rcond=None, full= False, w = np.divide(1., noise_generic_pc))								
            # CaT: Ca1 + Ca2 + Ca3
            # Ca1 central bandpass = [8484.0, 8513.0]
            flux_Ca1=y[np.logical_and(x>=8484.,x<=8513.)]
            lambda_Ca1=x[np.logical_and(x>=8484.,x<=8513.)]
            pc_Ca1=pc[0]*lambda_Ca1+pc[1]
            #	Ca1 = simps( 1. - ( flux_Ca1 / pc_Ca1 ), lambda_Ca1, even='avg' )
            Ca1=disp*np.sum(1.-(flux_Ca1/pc_Ca1))
            # Ca2 central bandpass = [8522.0, 8562.0]
            flux_Ca2=y[np.logical_and(x>=8522.,x<=8562.)]
            lambda_Ca2=x[np.logical_and(x>=8522.,x<=8562.)]
            pc_Ca2=pc[0]*lambda_Ca2+pc[1]
            #	Ca2 = simps( 1. - ( flux_Ca2 / pc_Ca2 ), lambda_Ca2, even='avg' )
            Ca2=disp*np.sum(1.-(flux_Ca2/pc_Ca2))
            # Ca3 central bandpass = [8642.0, 8682.0]
            flux_Ca3=y[np.logical_and(x>=8642.,x<=8682.)]
            lambda_Ca3=x[np.logical_and(x>=8642.,x<=8682.)]
            pc_Ca3=pc[0]*lambda_Ca3+pc[1]
            #	Ca3 = simps( 1. - ( flux_Ca3 / pc_Ca3 ), lambda_Ca3, even='avg' )
            Ca3=disp*np.sum(1.-(flux_Ca3/pc_Ca3))
            #print 'Ca', Ca1, Ca2, Ca3
            CaT=Ca1+Ca2+Ca3
            #print 'CaT', CaT
            # PaT: Pa1 + Pa2 + Pa3
            # Pa1 central bandpass = [8461.0, 8474.0]
            flux_Pa1=y[np.logical_and(x>=8461.,x<=8474.)]
            lambda_Pa1=x[np.logical_and(x>=8461.,x<=8474.)]
            pc_Pa1=pc[0]*lambda_Pa1+pc[1]
            #	Pa1 = simps( 1. - ( flux_Pa1 / pc_Pa1 ), lambda_Pa1, even='avg' )
            Pa1=disp*np.sum(1.-(flux_Pa1/pc_Pa1))
            # Pa2 central bandpass = [8577.0, 8619.0]
            flux_Pa2=y[np.logical_and(x>=8577.,x<=8619.)]
            lambda_Pa2=x[np.logical_and(x>=8577.,x<=8619.)]
            pc_Pa2=pc[0]*lambda_Pa2+pc[1]
            #	Pa2 = simps( 1. - ( flux_Pa2 / pc_Pa2 ), lambda_Pa2, even='avg' )
            Pa2=disp*np.sum(1.-(flux_Pa2/pc_Pa2))
            # Pa3 central bandpass = [8730.0, 8772.0]
            flux_Pa3=y[np.logical_and(x>=8730.,x<=8772.)]
            lambda_Pa3=x[np.logical_and(x>=8730.,x<=8772.)]
            pc_Pa3=pc[0]*lambda_Pa3+pc[1]
            #	Pa3 = simps( 1. - ( flux_Pa3 / pc_Pa3 ), lambda_Pa3, even='avg' )
            Pa3=disp*np.sum(1.-(flux_Pa3/pc_Pa3))
            PaT=Pa1+Pa2+Pa3
            #print 'PaT', PaT
            # CaT^* = CaT - 0.93 PaT (Cenarro et al. (2001)) 
            lick_value=CaT-(0.93*PaT) 
            # Error CaT^*:
            flux_SN_CaT=y[(x>=8484.)*(x<=8513.)+(x>=8522.)*(x<=8562.)+(x>=8642.)*(x<=8682.)+(x>=8461.)*(x<=8474.)+(x>=8577.)*(x<=8619.)+(x>=8730.)*(x<=8772.)+(x>=8474.)*(x<=8484.)+(x>=8562.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            noise_SN_CaT=erry[(x>=8484.)*(x<=8513.)+(x>=8522.)*(x<=8562.)+(x>=8642.)*(x<=8682.)+(x>=8461.)*(x<=8474.)+(x>=8577.)*(x<=8619.)+(x>=8730.)*(x<=8772.)+(x>=8474.)*(x<=8484.)+(x>=8562.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            rapp_SN_CaT=np.sum(np.divide(flux_SN_CaT,noise_SN_CaT))
            N_SN_CaT=len(flux_SN_CaT)
            SN_CaT=(1./(N_SN_CaT*np.sqrt(disp)))*rapp_SN_CaT
            lick_error=(16.43-(0.1052*lick_value))/SN_CaT	
            #	#xp = np.linspace(8400, 8800, 1200)
            #	_ = plt.plot(lambda_Pa1, flux_Pa1, '.-', lambda_Pa1, pc_Pa1, '--',lambda_Pa2, flux_Pa2, '.-', lambda_Pa2, pc_Pa2, '--',lambda_Pa3, flux_Pa3, '.-', lambda_Pa3, pc_Pa3, '--')
            #	plt.xlim(8350, 8900)
            #	#plt.ylim(0.0001, 0.0003)
            #	plt.show()
        else:
            lick_value=self.nan
            lick_error=self.nan
        return lick_value, lick_error

    def _measure_ind_general_wei(self,x,y,erry,disp):
        """
        .. warning:
        To be implemented
        """
        print(" This feature is not available yet.")

        if (x.min()<bb and x.max()>rr):
            # Ref. Cenarro et al. (2001)
            # pseudocontinuum:
            flux_generic_pc=y[(x>=8474.)*(x<=8484.)+(x>=8563.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            noise_generic_pc=erry[(x>=8474.)*(x<=8484.)+(x>=8563.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]		
            lambda_generic_pc=x[(x>=8474.)*(x<=8484.)+(x>=8563.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            pc = np.polyfit(lambda_generic_pc, flux_generic_pc, 1, rcond=None, full= False, w = np.divide(1., noise_generic_pc))								
            # CaT: Ca1 + Ca2 + Ca3
            # Ca1 central bandpass = [8484.0, 8513.0]
            flux_Ca1=y[np.logical_and(x>=8484.,x<=8513.)]
            lambda_Ca1=x[np.logical_and(x>=8484.,x<=8513.)]
            pc_Ca1=pc[0]*lambda_Ca1+pc[1]
            #	Ca1 = simps( 1. - ( flux_Ca1 / pc_Ca1 ), lambda_Ca1, even='avg' )
            Ca1=disp*np.sum(1.-(flux_Ca1/pc_Ca1))
            # Ca2 central bandpass = [8522.0, 8562.0]
            flux_Ca2=y[np.logical_and(x>=8522.,x<=8562.)]
            lambda_Ca2=x[np.logical_and(x>=8522.,x<=8562.)]
            pc_Ca2=pc[0]*lambda_Ca2+pc[1]
            #	Ca2 = simps( 1. - ( flux_Ca2 / pc_Ca2 ), lambda_Ca2, even='avg' )
            Ca2=disp*np.sum(1.-(flux_Ca2/pc_Ca2))
            # Ca3 central bandpass = [8642.0, 8682.0]
            flux_Ca3=y[np.logical_and(x>=8642.,x<=8682.)]
            lambda_Ca3=x[np.logical_and(x>=8642.,x<=8682.)]
            pc_Ca3=pc[0]*lambda_Ca3+pc[1]
            #	Ca3 = simps( 1. - ( flux_Ca3 / pc_Ca3 ), lambda_Ca3, even='avg' )
            Ca3=disp*np.sum(1.-(flux_Ca3/pc_Ca3))
            #print 'Ca', Ca1, Ca2, Ca3
            CaT=Ca1+Ca2+Ca3
            #print 'CaT', CaT
            # PaT: Pa1 + Pa2 + Pa3
            # Pa1 central bandpass = [8461.0, 8474.0]
            flux_Pa1=y[np.logical_and(x>=8461.,x<=8474.)]
            lambda_Pa1=x[np.logical_and(x>=8461.,x<=8474.)]
            pc_Pa1=pc[0]*lambda_Pa1+pc[1]
            #	Pa1 = simps( 1. - ( flux_Pa1 / pc_Pa1 ), lambda_Pa1, even='avg' )
            Pa1=disp*np.sum(1.-(flux_Pa1/pc_Pa1))
            # Pa2 central bandpass = [8577.0, 8619.0]
            flux_Pa2=y[np.logical_and(x>=8577.,x<=8619.)]
            lambda_Pa2=x[np.logical_and(x>=8577.,x<=8619.)]
            pc_Pa2=pc[0]*lambda_Pa2+pc[1]
            #	Pa2 = simps( 1. - ( flux_Pa2 / pc_Pa2 ), lambda_Pa2, even='avg' )
            Pa2=disp*np.sum(1.-(flux_Pa2/pc_Pa2))
            # Pa3 central bandpass = [8730.0, 8772.0]
            flux_Pa3=y[np.logical_and(x>=8730.,x<=8772.)]
            lambda_Pa3=x[np.logical_and(x>=8730.,x<=8772.)]
            pc_Pa3=pc[0]*lambda_Pa3+pc[1]
            #	Pa3 = simps( 1. - ( flux_Pa3 / pc_Pa3 ), lambda_Pa3, even='avg' )
            Pa3=disp*np.sum(1.-(flux_Pa3/pc_Pa3))
            PaT=Pa1+Pa2+Pa3
            #print 'PaT', PaT
            # CaT^* = CaT - 0.93 PaT (Cenarro et al. (2001)) 
            lick_value=CaT-(0.93*PaT) 
            # Error CaT^*:
            flux_SN_CaT=y[(x>=8484.)*(x<=8513.)+(x>=8522.)*(x<=8562.)+(x>=8642.)*(x<=8682.)+(x>=8461.)*(x<=8474.)+(x>=8577.)*(x<=8619.)+(x>=8730.)*(x<=8772.)+(x>=8474.)*(x<=8484.)+(x>=8562.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            noise_SN_CaT=erry[(x>=8484.)*(x<=8513.)+(x>=8522.)*(x<=8562.)+(x>=8642.)*(x<=8682.)+(x>=8461.)*(x<=8474.)+(x>=8577.)*(x<=8619.)+(x>=8730.)*(x<=8772.)+(x>=8474.)*(x<=8484.)+(x>=8562.)*(x<=8577.)+(x>=8619.)*(x<=8642.)+(x>=8700.)*(x<=8725.)+(x>=8776.)*(x<=8792.)]
            rapp_SN_CaT=np.sum(np.divide(flux_SN_CaT,noise_SN_CaT))
            N_SN_CaT=len(flux_SN_CaT)
            SN_CaT=(1./(N_SN_CaT*np.sqrt(disp)))*rapp_SN_CaT
            lick_error=(16.43-(0.1052*lick_value))/SN_CaT	
            #	#xp = np.linspace(8400, 8800, 1200)
            #	_ = plt.plot(lambda_Pa1, flux_Pa1, '.-', lambda_Pa1, pc_Pa1, '--',lambda_Pa2, flux_Pa2, '.-', lambda_Pa2, pc_Pa2, '--',lambda_Pa3, flux_Pa3, '.-', lambda_Pa3, pc_Pa3, '--')
            #	plt.xlim(8350, 8900)
            #	#plt.ylim(0.0001, 0.0003)
            #	plt.show()
        else:
            lick_value=self.nan
            lick_error=self.nan
        return lick_value, lick_error

        #-------------------------------------------------------------


    # ---------------------------------------------------------
    def __init__(self, wave, flux, ferr, todo, regions, units, names, meas_method, BPR_thres, nans, verbose):

        Nindices         = len(regions)
        self.wave        = wave
        self.flux        = flux
        self.ferr        = ferr 
        self.todo        = todo
        self.regions     = regions
        self.units       = units
        self.nan         = nans
        self.int_binsize = 0.0625
        self.intp_kind   = 'linear'
        self.debug       = verbose

        # Average dispersion of the spectrum
        # SPECTRA SHOULD BE LINEARLY SAMPLED, so disp=pixel dimension
        dispv     = np.diff(self.wave)
        self.disp = dispv.mean()

        self.vals = np.full(Nindices, np.nan, dtype=float)
        self.errs = np.full(Nindices, np.nan, dtype=float)
        self.BPR  = np.full(Nindices, np.nan, dtype=float)

        # if meas_method == 'int':
        #     # Interpolate spectrum (0.0625 AA)
        #     self._flux_interp(bin=0.0625)
        
        if wave[todo].sum() > 10:    # To avoid errors due to very bad spactra while running the catalog

            for k in range(Nindices):

                self._check_ind_def(self.regions[k],self.units[k])   # To move into the indices module


                if meas_method == 'exact':	
                    # self.flux_intp = interpolate.interp1d(wave[todo],flux[todo], kind='zero')
                    # self.ferr_intp  = interpolate.interp1d(wave[todo],ferr[todo], kind='zero')
                    # # self.flux_intp = lambda x: np.interp(x, wave, flux)
                    # # self.ferr_intp  = lambda x: np.interp(x, wave, ferr)
                    bb, rr = np.nanmin(wave[todo]), np.nanmax(wave[todo])
                    self._flux_interp_exact(bb,rr,binsize=self.int_binsize)

                    self.vals[k], self.errs[k], self.BPR[k] = self._measure_multi_exact(self.regions[k],self.units[k])


                elif meas_method == 'int':
                    self.vals[k], self.errs[k], self.BPR[k] = self._measure_multi_interp(self.regions[k],self.units[k])


                elif meas_method == 'wei':
                    self.vals[k], self.errs[k], self.BPR[k] = self._measure_multi_wei(self.regions[k],self.units[k])
                    

                else:
                    raise ValueError("PyLick: meas_method not understood.")
        
        bad             = self.BPR > BPR_thres
        self.vals[bad]  = self.nan
        self.errs[bad]  = self.nan

        self.finite = np.logical_and(np.isfinite(self.vals), np.isfinite(self.errs))
